/* autor : Max Celedon Collins
  Psycho World
  http://www.psychoworld.cl
  */

import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import javax.microedition.lcdui.game.*;
import java.util.*;
import java.io.*;

public class Spectrum extends Z80 {
	public Graphics		parentGraphics = null;
	public Graphics		canvasGraphics = null;
	public Graphics		bufferGraphics = null;
	public Image		bufferImage = null;

	public MobileZX parent      = null;

	private int gwidth, gheight;

	public Spectrum( MobileZX _parent) throws Exception {
		super(4); // 4 Mhz (J2ME and handsets in general don't have support for floating point numbers)

		parent = _parent;

		//pixel scale expresed as a ratio, default 1:1
		bufferImage = Image.createImage( (nPixelsWide*pixelScale)*5/5, (nPixelsHigh*pixelScale)*3/3 );
		
		bufferGraphics = bufferImage.getGraphics();
		canvasGraphics = getGraphics();
		gwidth = bufferImage.getWidth();
		gheight = bufferImage.getHeight();
	}

	public int inb( int port ) {
		int res = 0xff;

		if ( (port & 0x0001) == 0 ) {
			if ( (port & 0x8000) == 0 ) { res &= _B_SPC; }
			if ( (port & 0x4000) == 0 ) { res &= _H_ENT; }
			if ( (port & 0x2000) == 0 ) { res &= _Y_P;   }
			if ( (port & 0x1000) == 0 ) { res &= _6_0;   }
			if ( (port & 0x0800) == 0 ) { res &= _1_5;   }
			if ( (port & 0x0400) == 0 ) { res &= _Q_T;   }
			if ( (port & 0x0200) == 0 ) { res &= _A_G;   }
			if ( (port & 0x0100) == 0 ) { res &= _CAPS_V;}
		}

		return(res);
	}
  
	public void outb( int port, int outByte, int tstates ) {
		if ( (port & 0x0001) == 0 ) {
			newBorder = (outByte & 0x07);
		}
	}

	public void pokeb( int addr, int newByte ) {
		if ( addr >= (22528+768) ) {
			mem[ addr ] = newByte;
			return;
		}

		if ( addr < 16384 ) {
			return;
		}

		if ( mem[ addr ] != newByte ) {
			plot( addr, newByte );
			mem[ addr ] = newByte;
		}
	}

	public void pokew( int addr, int word ) {
		int _mem[] = mem;

		if ( addr >= (22528+768) ) {
			_mem[ addr ] = word & 0xff;
			if ( ++addr != 65536 ) {
				_mem[ addr ] = word >> 8;
			}
			return;
		}

		if ( addr < 16384 ) {
			return;
		}

		int        newByte0 = word & 0xff;
		if ( _mem[ addr ] != newByte0 ) {
			plot( addr, newByte0 );
			_mem[ addr ] = newByte0;
		}

		int        newByte1 = word >> 8;
		if ( ++addr != (22528+768) ) {
			if ( _mem[ addr ] != newByte1 ) {
				plot( addr, newByte1 );
				_mem[ addr ] = newByte1;
			}
		}
		else {
			_mem[ addr ] = newByte1;
		}
	}
	public  int     sleepHack = 0;
	public  int     refreshRate = 5;  // refresh screen each 'n' interrupts
	private int     interruptCounter = 0;
	private boolean resetAtNextInterrupt = false;
	private boolean pauseAtNextInterrupt = false;
	private boolean refreshNextInterrupt = true;
	private boolean loadFromURLFieldNextInterrupt = false;

	public  Thread  pausedThread = null;
	public  long    timeOfLastInterrupt = 0;
	private long    timeOfLastSample = 0;

	public final int interrupt() {

		if ( resetAtNextInterrupt ) {
			resetAtNextInterrupt = false;
			reset();
		}

		interruptCounter++;

		// flash characters each a half of second
		if ( (interruptCounter % 25) == 0 ) {
			refreshFlashChars();
		}

		if ( (interruptCounter % refreshRate) == 0 ) {
			screenPaint();
			resetKeyboard();
		}

		checkCelularKeyboard();
		return super.interrupt();
	}

	public void reset() {
		super.reset();
		outb( 254, 0xff, 0 );
	}

	public  final int pixelScale  = 1;    // pixel scale
	public final int nPixelsWide = 256;
	public final int nPixelsHigh = 192;
	public final int nCharsWide  = 32;
	public final int nCharsHigh  = 24;

	private final int sat = 238;
	private final int sat2 = 255;

	private final Color brightColors[] = {
		new Color(   0,   0,   0 ),  new Color(   0,   0, sat ),
		new Color( sat,   0,   0 ),  new Color( sat,   0, sat ),
		new Color(   0, sat,   0 ),  new Color(   0, sat, sat ),
		new Color( sat, sat,   0 ),  new Color( sat, sat, sat ),
		new Color(   0,   0,   0 ),  new Color(   0,   0, sat2 ),
		new Color( sat2,   0,   0 ),  new Color( sat2,   0, sat2 ),
		new Color(   0, sat2,   0 ),  new Color(   0, sat2, sat2 ),
		new Color( sat2, sat2,   0 ),  new Color( sat2, sat2, sat2 )
	};

	private final int firstAttr = (nPixelsHigh*nCharsWide);
	private final int lastAttr  = firstAttr + (nCharsHigh*nCharsWide);

	private int first = -1;
	private int FIRST = -1;
	private final int last[] = new int[ (nPixelsHigh+nCharsHigh)*nCharsWide ];
	private final int next[] = new int[ (nPixelsHigh+nCharsHigh)*nCharsWide ];

	public int newBorder = 7; 
	private boolean flashInvert = false;

	private final void refreshFlashChars() {
		flashInvert = !flashInvert;
		for ( int i = firstAttr; i < lastAttr; i++ ) {
			int attr = mem[i+16384];
			if ( (attr & 0x80) != 0 ) {
				last[i] = (~attr) & 0xff;
				if ( next[i] == -1 ) {
					next[i] = FIRST;
					FIRST = i;
				}
			}
		}
	}

	public final void refreshWholeScreen() {

		for ( int i = 0; i < firstAttr; i++ ) {
			next[ i ] = i-1;
			last[ i ] = (~mem[ i+16384 ]) & 0xff;
		}

		for ( int i = firstAttr; i < lastAttr; i++ ) {
			next[ i ] = -1;
			last[ i ] = mem[ i+16384 ];
		}

		first = firstAttr - 1;
		FIRST = -1;
	}

	private final void plot( int addr, int newByte ) {
		int     offset = addr - 16384;

		if ( next[ offset ] == -1 ) {
			if ( offset < firstAttr ) {
				next[ offset ] = first;
				first = offset;
			}
			else {
				next[ offset ] = FIRST;
				FIRST = offset;
			}
		}
	}

 	public boolean	runAtFullSpeed = true;

	public  synchronized Image getImage( int attr, int pattern ) {
		try {
			return tryGetImage( attr, pattern );
		}
		catch ( OutOfMemoryError e ) {
			imageMap   = null;
			patternMap = null;

			System.gc();
			patternMap = new Hashtable();
			imageMap   = new Image[ 1<<11 ];

			return tryGetImage(  attr, pattern );
		}
	}

	public Hashtable patternMap = new Hashtable();
	public Image imageMap[] = new Image[ 1<<11 ]; // 7 bits for attribute, 4 bits for patterns

	private  Image tryGetImage(  int attr, int pattern ) {
		int bright = ((attr>>3) & 0x08);
		int ink = ((attr   ) & 0x07) | bright;
		int pap = ((attr>>3) & 0x07) | bright;
		int hashValue = 0;

		for ( int i = 0; i < 4; i++ ) {
			int col = ((pattern & (1<<i)) == 0) ? pap : ink;
			hashValue |= (col << (i<<2));
		}

		Integer	imageKey = new Integer( hashValue );
		Image image = (Image) patternMap.get( imageKey );
		if ( image == null ) {
			Color	colors[] = brightColors;

			image = Image.createImage( 4, 1 );
			Graphics g = image.getGraphics();

			for ( int i = 0; i < 4; i++ ) {
				int col = ((pattern & (1<<i)) == 0) ? pap : ink;
				g.setColor( colors[ col ].R, colors[ col ].G, colors[ col ].B  );
				g.fillRect( (3-i), 0, 1, 1 );
			}

			patternMap.put( imageKey, image );
		}

		return image;
	}

	public final void screenPaint() {
		int addr   = FIRST;

		// Update pixels affected by the changes in the attribute table
		while ( addr >= 0 ) {
			int        oldAttr = last[ addr ];
			int        newAttr = mem[ addr + 16384 ];
			last[ addr ] = newAttr;

			boolean    inkChange    = ((oldAttr & 0x47) != (newAttr & 0x47));
			boolean    papChange    = ((oldAttr & 0x78) != (newAttr & 0x78));
			boolean    flashChange  = ((oldAttr & 0x80) != (newAttr & 0x80));

			if ( inkChange || papChange || flashChange ) {
				boolean    allChange = ((inkChange && papChange) || flashChange);
				int        scrAddr   = ((addr & 0x300) << 3) | (addr & 0xff);

				for ( int i = 8; i != 0; i-- ) {

					if ( allChange ) {
						last[ scrAddr ] = ((~mem[ scrAddr+16384 ]) & 0xff);
					}
					else {
						int	oldPixels = last[ scrAddr ];
						int	newPixels = mem[ scrAddr+16384 ];
						int	changes = oldPixels ^ newPixels;

						if ( inkChange ) {
							changes |=  newPixels;
						}
						else {
							changes |= ((~newPixels) & 0xff);
						}
						if ( changes == 0 ) {
							scrAddr += 256;
							continue;
						}
						last[ scrAddr ] = changes ^ newPixels;
					}

					if ( next[ scrAddr ] == -1 ) {
						next[ scrAddr ] = first;
						first = scrAddr;
					}

					scrAddr += 256;
				}
			}

			int        newAddr = next[ addr ];
			next[ addr ] = -1;
			addr = newAddr;
		}
		FIRST = -1;

		// Update screen only if neccesary
		if ( first < 0 ) {
			return;
		}
		// Update affected pixels
		addr     = first;
		while ( addr >= 0 ) {
			int oldPixels = last[ addr ];
			int newPixels = mem[ addr+16384 ];
			int changes   = oldPixels ^ newPixels;
			last[ addr ] = newPixels;

			int x = ((addr&0x1f) << 3);
			int y = (((int)(addr&0x00e0))>>2) +
				(((int)(addr&0x0700))>>8) +
				(((int)(addr&0x1800))>>5);
			int X = (x*pixelScale*5/5);
			int Y = (y*pixelScale*3/3);

			int attr = mem[ 22528 + (addr&0x1f) + ((y>>3)*nCharsWide) ];

			// exchange colors if flash is on
			if ( flashInvert && ((attr & 0x80) != 0) ) {
				newPixels = (~newPixels & 0xff);
			}

			// Redraw left pattern if neccesary
			if ( (changes & 0xf0) != 0 ) {
				int newPixels1 = (newPixels&0xf0)>>4;
				int imageMapEntry1 = (((attr & 0x7f)<<4) | newPixels1);
				Image image1 = imageMap[ imageMapEntry1 ];
				if ( image1 == null ) {
					image1 = getImage(  attr, newPixels1 );
					imageMap[ imageMapEntry1 ] = image1;
				}
				bufferGraphics.drawImage( image1, X, Y, Graphics.TOP|Graphics.LEFT );
			}

			// Redraw right pattern if neccesary
			if ( (changes & 0x0f) != 0 ) {
				int newPixels2 = (newPixels&0x0f);
				int imageMapEntry2 = (((attr & 0x7f)<<4) | newPixels2);
				Image image2 = imageMap[ imageMapEntry2 ];
				if ( image2 == null ) {
					image2 = getImage(  attr, newPixels2 );
					imageMap[ imageMapEntry2 ] = image2;
				}
				bufferGraphics.drawImage( image2, X+4, Y, Graphics.TOP|Graphics.LEFT );
			}

			int        newAddr = next[ addr ];
			next[ addr ] = -1;
			addr = newAddr;
		}
		first = -1;

		paintBuffer();
	}

	public void paintBuffer() {
		canvasGraphics.drawImage( bufferImage, 0, 0, Graphics.TOP|Graphics.LEFT );
		flushGraphics(0,0,gwidth, gheight);
	}

	private final int b4 = 0x10;
	private final int b3 = 0x08;
	private final int b2 = 0x04;
	private final int b1 = 0x02;
	private final int b0 = 0x01;

	private int _B_SPC  = 0xff;
	private int _H_ENT  = 0xff;
	private int _Y_P    = 0xff;
	private int _6_0    = 0xff;
	private int _1_5    = 0xff;
	private int _Q_T    = 0xff;
	private int _A_G    = 0xff;
	private int _CAPS_V = 0xff;

	public void resetKeyboard() {
		_B_SPC  = 0xff;
		_H_ENT  = 0xff;
		_Y_P    = 0xff;
		_6_0    = 0xff;
		_1_5    = 0xff;
		_Q_T    = 0xff;
		_A_G    = 0xff;
		_CAPS_V = 0xff;
	}

	public void checkCelularKeyboard(){
		 int keyState = getKeyStates();
		 boolean down = true;
//		 boolean CAPS = true;
//		 boolean SHIFT=true;
//		 boolean SYMB=true;

		 if ((keyState & LEFT_PRESSED) != 0) {
			 K5( down );
		}
 		else if ((keyState & RIGHT_PRESSED) != 0) {
 			 K8( down );
 		}
		else if ((keyState & UP_PRESSED) != 0) {
 			K7( down );
 		}
		else if ((keyState & DOWN_PRESSED) != 0) {
 			K6( down );
				System.out.println("down");
 		}
		else if ((keyState & FIRE_PRESSED) != 0) {
 			KSPC( down );
				System.out.println("FIRE");
 		}
 		else if ((keyState & KEY_NUM0) != 0) {
 				K0( down );
				System.out.println("key 0");
 		}
		else if ((keyState & KEY_NUM1) != 0) {
 				K1( down );
 		}
		else if ((keyState & KEY_NUM2) != 0) {
 				K2( down );
 		}
		else if ((keyState & KEY_NUM3) != 0) {
 				K3( down );
 		}
		else if ((keyState & KEY_NUM4) != 0) {
 				K4( down );
 		}
		else if ((keyState & KEY_NUM5) != 0) {
 				K5( down );
 		}
		else if ((keyState & KEY_NUM6) != 0) {
 				K6( down );
 		}
		else if ((keyState & KEY_NUM7) != 0) {
 				K7( down );
 		}
		else if ((keyState & KEY_NUM8) != 0) {
 				K8( down );
 		}
		else if ((keyState & KEY_NUM9) != 0) {
 				K9( down );
 		}
	}



	private final void K1( boolean down ) {
		if ( down ) _1_5 &= ~b0; else _1_5 |= b0;
	}
	private final void K2( boolean down ) {
		if ( down ) _1_5 &= ~b1; else _1_5 |= b1;
	}
	private final void K3( boolean down ) {
		if ( down ) _1_5 &= ~b2; else _1_5 |= b2;
	}
	private final void K4( boolean down ) {
		if ( down ) _1_5 &= ~b3; else _1_5 |= b3;
	}
	private final void K5( boolean down ) {
		if ( down ) _1_5 &= ~b4; else _1_5 |= b4;
	}

	private final void K6( boolean down ) {
		if ( down ) _6_0 &= ~b4; else _6_0 |= b4;
	}
	private final void K7( boolean down ) {
		if ( down ) _6_0 &= ~b3; else _6_0 |= b3;
	}
	private final void K8( boolean down ) {
		if ( down ) _6_0 &= ~b2; else _6_0 |= b2;
	}
	private final void K9( boolean down ) {
		if ( down ) _6_0 &= ~b1; else _6_0 |= b1;
	}
	private final void K0( boolean down ) {
		if ( down ) _6_0 &= ~b0; else _6_0 |= b0;
	}


	private final void KQ( boolean down ) {
		if ( down ) _Q_T &= ~b0; else _Q_T |= b0;
	}
	private final void KW( boolean down ) {
		if ( down ) _Q_T &= ~b1; else _Q_T |= b1;
	}
	private final void KE( boolean down ) {
		if ( down ) _Q_T &= ~b2; else _Q_T |= b2;
	}
	private final void KR( boolean down ) {
		if ( down ) _Q_T &= ~b3; else _Q_T |= b3;
	}
	private final void KT( boolean down ) {
		if ( down ) _Q_T &= ~b4; else _Q_T |= b4;
	}

	private final void KY( boolean down ) {
		if ( down ) _Y_P &= ~b4; else _Y_P |= b4;
	}
	private final void KU( boolean down ) {
		if ( down ) _Y_P &= ~b3; else _Y_P |= b3;
	}
	private final void KI( boolean down ) {
		if ( down ) _Y_P &= ~b2; else _Y_P |= b2;
	}
	private final void KO( boolean down ) {
		if ( down ) _Y_P &= ~b1; else _Y_P |= b1;
	}
	private final void KP( boolean down ) {
		if ( down ) _Y_P &= ~b0; else _Y_P |= b0;
	}


	private final void KA( boolean down ) {
		if ( down ) _A_G &= ~b0; else _A_G |= b0;
	}
	private final void KS( boolean down ) {
		if ( down ) _A_G &= ~b1; else _A_G |= b1;
	}
	private final void KD( boolean down ) {
		if ( down ) _A_G &= ~b2; else _A_G |= b2;
	}
	private final void KF( boolean down ) {
		if ( down ) _A_G &= ~b3; else _A_G |= b3;
	}
	private final void KG( boolean down ) {
		if ( down ) _A_G &= ~b4; else _A_G |= b4;
	}

	private final void KH( boolean down ) {
		if ( down ) _H_ENT &= ~b4; else _H_ENT |= b4;
	}
	private final void KJ( boolean down ) {
		if ( down ) _H_ENT &= ~b3; else _H_ENT |= b3;
	}
	private final void KK( boolean down ) {
		if ( down ) _H_ENT &= ~b2; else _H_ENT |= b2;
	}
	private final void KL( boolean down ) {
		if ( down ) _H_ENT &= ~b1; else _H_ENT |= b1;
	}
	private final void KENT( boolean down ) {
		if ( down ) _H_ENT &= ~b0; else _H_ENT |= b0;
	}


	private final void KCAPS( boolean down ) {
		if ( down ) _CAPS_V &= ~b0; else _CAPS_V |= b0;
	}
	private final void KZ( boolean down ) {
		if ( down ) _CAPS_V &= ~b1; else _CAPS_V |= b1;
	}
	private final void KX( boolean down ) {
		if ( down ) _CAPS_V &= ~b2; else _CAPS_V |= b2;
	}
	private final void KC( boolean down ) {
		if ( down ) _CAPS_V &= ~b3; else _CAPS_V |= b3;
	}
	private final void KV( boolean down ) {
		if ( down ) _CAPS_V &= ~b4; else _CAPS_V |= b4;
	}

	private final void KB( boolean down ) {
		if ( down ) _B_SPC &= ~b4; else _B_SPC |= b4;
	}
	private final void KN( boolean down ) {
		if ( down ) _B_SPC &= ~b3; else _B_SPC |= b3;
	}
	private final void KM( boolean down ) {
		if ( down ) _B_SPC &= ~b2; else _B_SPC |= b2;
	}
	private final void KSYMB( boolean down ) {
		if ( down ) _B_SPC &= ~b1; else _B_SPC |= b1;
	}
	private final void KSPC( boolean down ) {
		if ( down ) _B_SPC &= ~b0; else _B_SPC |= b0;
	}

	public final boolean doKey( boolean down, int ascii, int mods ) {
		return true;
	}


	public void loadSnapshot( String name, InputStream is, int snapshotLength ) throws Exception {
		
		// Crude check but it'll work (SNA is a fixed size)
		if ( (snapshotLength == 49179) ) {
			loadSNA( name, is );
		}
		else {
			loadZ80( name, is, snapshotLength );
		}

		refreshWholeScreen();
		resetKeyboard();
	}

	public void loadROM( String name, java.io.InputStream is ) throws Exception {
		readBytes( is, mem, 0, 16384 );
	}

	public void loadSNA( String name, java.io.InputStream is ) throws Exception {
		int        header[] = new int[27];
		readBytes( is, header, 0,        27 );
		System.out.println("lectura header exitosa");
		readBytes( is, mem,    16384, 49152 );
		System.out.println("lectura cuerpo exitosa");

		I( header[0] );

		HL( header[1] | (header[2]<<8) );
		DE( header[3] | (header[4]<<8) );
		BC( header[5] | (header[6]<<8) );
		AF( header[7] | (header[8]<<8) );

		exx();
		ex_af_af();

		HL( header[9]  | (header[10]<<8) );
		DE( header[11] | (header[12]<<8) );
		BC( header[13] | (header[14]<<8) );

		IY( header[15] | (header[16]<<8) );
		IX( header[17] | (header[18]<<8) );

		if ( (header[19] & 0x04)!= 0 ) {
			IFF2( true );
		}
		else {
			IFF2( false );
		}

		R( header[20] );

		AF( header[21] | (header[22]<<8) );
		SP( header[23] | (header[24]<<8) );

		switch( header[25] ) {
		case 0:
			IM( IM0 );
			break;
		case 1:
			IM( IM1 );
			break;
		default:
			IM( IM2 );
			break;
		}

		outb( 254, header[26], 0 ); // border

		IFF1( IFF2() );
		REFRESH( 2 );
		poppc();

	}

	public void loadZ80( String name,  java.io.InputStream is, int bytesLeft ) throws Exception {
		System.out.println("bytesleft="+bytesLeft);

		int        header[] = new int[30];
		boolean    compressed = false;

		bytesLeft -= readBytes( is, header, 0, 30 );

		System.out.println("bytesleft="+bytesLeft);

		A( header[0] );
		F( header[1] );

		C( header[2] );
		B( header[3] );
		L( header[4] );
		H( header[5] );

		PC( header[6] | (header[7]<<8) );
		SP( header[8] | (header[9]<<8) );

		I( header[10] );
		R( header[11] );

		int tbyte = header[12];
		if ( tbyte == 255 ) {
			tbyte = 1;
		}

		outb( 254, ((tbyte >> 1) & 0x07), 0 ); // border

		if ( (tbyte & 0x01) != 0 ) {
			R( R() | 0x80 );
		}
		compressed = ((tbyte & 0x20) != 0);

		E( header[13] );
		D( header[14] );

		ex_af_af();
		exx();

		C( header[15] );
		B( header[16] );
		E( header[17] );
		D( header[18] );
		L( header[19] );
		H( header[20] );

		A( header[21] );
		F( header[22] );

		ex_af_af();
		exx();

		IY( header[23] | (header[24]<<8) );
		IX( header[25] | (header[26]<<8) );

		IFF1( header[27] != 0 );
		IFF2( header[28] != 0 );

		switch ( header[29] & 0x03 ) {
		case 0:
			IM( IM0 );
			break;
		case 1:
			IM( IM1 );
			break;
		default:
			IM( IM2 );
			break;
		}

		if ( PC() == 0 ) {
			System.out.println("bytesleft="+bytesLeft);
			loadZ80_extended( is, bytesLeft );
			return;
		}


		if ( compressed ) {
			int data[] = new int[ bytesLeft ];
			int addr   = 16384;

			int size = readBytes( is, data, 0, bytesLeft );
			int i    = 0;

			while ( (addr < 65536) && (i < size) ) {
				tbyte = data[i++];
				if ( tbyte != 0xed ) {
					pokeb( addr, tbyte );
					addr++;
				}
				else {
					tbyte = data[i++];
					if ( tbyte != 0xed ) {
						pokeb( addr, 0xed );
						i--;
						addr++;
					}
					else {
						int        count;
						count = data[i++];
						tbyte = data[i++];
						while ( (count--) != 0 ) {
							pokeb( addr, tbyte );
							addr++;
						}
					}
				}
			}
		}
		else {
			readBytes( is, mem, 16384, 49152 );
		}
	}

	private void loadZ80_extended( InputStream is, int bytesLeft ) throws Exception {
		int header[] = new int[2];

		System.out.println("bytesleft="+bytesLeft);

		bytesLeft -= readBytes( is, header, 0, header.length );

		int type = header[0] | (header[1] << 8);

		switch( type ) {
		case 23: 
			loadZ80_v201( is, bytesLeft );
			break;
		case 54: 
			loadZ80_v300( is, bytesLeft );
			break;
		case 58: 
			loadZ80_v301( is, bytesLeft );
			break;
		default:
			throw new Exception( "Z80 (extended): unsupported type " + type );
		}
	}

	private void loadZ80_v201( InputStream is, int bytesLeft ) throws Exception {
		int header[] = new int[23];
		bytesLeft -= readBytes( is, header, 0, header.length );

		PC( header[0] | (header[1]<<8) );

		int type = header[2];

		if ( type > 1 ) {
			throw new Exception( "Z80 (v201): unsupported type " + type );
		}

		int data[] = new int[ bytesLeft ];
		readBytes( is, data, 0, bytesLeft );

		for ( int offset = 0, j = 0; j < 3; j++ ) {
			offset = loadZ80_page( data, offset );
		}
	}

	private void loadZ80_v300( InputStream is, int bytesLeft ) throws Exception {
		int        header[] = new int[54];
		System.out.println ("paso 0 "+bytesLeft);
		bytesLeft -= readBytes( is, header, 0, header.length );

		PC( header[0] | (header[1]<<8) );
		int type = header[2];

		if ( type > 6 ) {
			throw new Exception( "Z80 (v300): unsupported type " + type );
		}
		int [] data = new int [bytesLeft];
		readBytes( is, data, 0, bytesLeft );

		for ( int offset = 0, j = 0; j < 3; j++ ) {
			offset = loadZ80_page( data, offset );
		}
	}

	private void loadZ80_v301( InputStream is, int bytesLeft ) throws Exception {
		int        header[] = new int[58];
		bytesLeft -= readBytes( is, header, 0, header.length );

		PC( header[0] | (header[1]<<8) );
		int type = header[2];

		if ( type > 7 ) {
			throw new Exception( "Z80 (v301): unsupported type " + type );
		}

		int data[] = new int[ bytesLeft ];
		readBytes( is, data, 0, bytesLeft );

		for ( int offset = 0, j = 0; j < 3; j++ ) {
			offset = loadZ80_page( data, offset );
		}
	}

	private int loadZ80_page( int data[], int i ) throws Exception {
		int blocklen;
		int page;

		blocklen  = data[i++];
		blocklen |= (data[i++]) << 8;
		page = data[i++];

		int addr;
		switch(page) {
		case 4:
			addr = 32768;
			break;
		case 5:
			addr = 49152;
			break;
		case 8:
			addr = 16384;
			break;
		default:
			throw new Exception( "Z80 (page): out of range " + page );
		}

		int        k = 0;
		while (k < blocklen) {
			int        tbyte = data[i++]; k++;
			if ( tbyte != 0xed ) {
				pokeb(addr, ~tbyte);
				pokeb(addr, tbyte);
				addr++;
			}
			else {
				tbyte = data[i++]; k++;
				if ( tbyte != 0xed ) {
					pokeb(addr, 0);
					pokeb(addr, 0xed);
					addr++;
					i--; k--;
				}
				else {
					int        count;
					count = data[i++]; k++;
					tbyte = data[i++]; k++;
					while ( count-- > 0 ) {
						pokeb(addr, ~tbyte);
						pokeb(addr, tbyte);
						addr++;
					}
				}
			}
		}

		if ((addr & 16383) != 0) {
			throw new Exception( "Z80 (page): overrun" );
		}

		return i;
	}


	public int bytesReadSoFar = 0;
	public int bytesToReadTotal = 0;

	private int readBytes( java.io.InputStream is, int a[], int off, int n ) throws Exception {
		try {
			byte [] buff = new byte[ n ];
			System.out.println("Entering to readbytes");
			int toRead = n;
			int nRead = is.read( buff, 0, n );

			System.out.println(nRead+" bytes read");

			for ( int i = 0; i < n; i++ ) {
				a[ i+off ] = (buff[i]+256)&0xff;
			}

			return nRead;
		}
		catch ( Exception e ) {
			System.err.println( e );
			e.printStackTrace();
			throw e;
		}
	}
}
